#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>

#include <lwip/opt.h>
#include <lwip/err.h>
#include <lwip/arch.h>
#include <lwip/debug.h>
#include <lwip/sys.h>

#include <lwip/stats.h>
#include <lwip/init.h>
#include <lwip/tcpip.h>
#include <lwip/netif.h>
#include <lwip/snmp.h>

#if LWIP_IPV6
#include <lwip/ethip6.h>
#endif

#include "./pktdrvr.h"
#include "./pkt_if.h"

#if ETH_PAD_SIZE
#error ETH_PAD_SIZE must be 0.
#endif

#define DO_CHANGE_MY_MAC  0

#if 1
  #define INITIAL_RX_MODE   PDRX_ALL_PACKETS
#else
  #define INITIAL_RX_MODE   PDRX_MULTICAST2
#endif

static err_t low_level_output (struct netif *netif, struct pbuf *p);

static PKT_STAT initial_stat;


/* Our main interface-driver intialisation routine. Called by netif_add()
 * to setup an new interface.
 */
err_t
pkt_init (struct netif *netif)
{
  static int if_index = 0;
  int        if_flags;
#if 0
  struct pkt_info2 *pi = calloc (sizeof(*pi),1);

  netif->state = (void*) pi;   /* use this to store instance data */
  pinfo->rx_mode = INITIAL_RX_MODE;
  PktInitDriver (pi->rx_mode);
#endif

  netif->name[0] = 'e';     /* our interfaces are named "e0", "e1"... */
  netif->name[1] = (char)('0' + if_index++);
  netif->output     = etharp_output;
  netif->linkoutput = low_level_output;
#if LWIP_IPV6
  netif->output_ip6 = ethip6_output;
#endif

#if LWIP_NETIF_HOSTNAME
  /* Initialize interface hostname */
  netif_set_hostname (netif, "lwip");
#endif

  if_flags = NETIF_FLAG_BROADCAST |
             NETIF_FLAG_ETHARP |
             NETIF_FLAG_IGMP |
             NETIF_FLAG_LINK_UP |
             NETIF_FLAG_ETHERNET;

  netif->mtu    = 1500;
  netif->flags  = if_flags;
  netif->hwaddr_len = ETHARP_HWADDR_LEN;

  NETIF_INIT_SNMP (netif, snmp_ifType_ethernet_csmacd, 100000000);

  /* Init the packet-driver with correct receive mode.
   */
  if (!PktInitDriver(INITIAL_RX_MODE))
     return (ERR_IF);

  if (PktGetStatistics(pktInfo.handle))
       memcpy (&initial_stat, &pktStat, sizeof(initial_stat));
  else memset (&initial_stat, 0, sizeof(initial_stat));

  PktQueueBusy (0);

  memcpy (&netif->hwaddr, &myAddress, sizeof(netif->hwaddr));

#if DO_CHANGE_MY_MAC
  /*
   * Change the MAC address to a unique value
   * so that multiple iface instances on the same packet-driver are
   * supported. This will most likely fail under Windows (with the
   * NDIS3KT or SwsVpkt drivers).
   */
  netif->hwaddr [ETHARP_HWADDR_LEN-1] += if_index;
  if (!PktSetAddress((const ETHER*)&netif->hwaddr))
  {
    LWIP_TRACE (1, "Failed to set ether address\n");
    netif->hwaddr [ETHARP_HWADDR_LEN-1] -= if_index;
 // pi->mac_changed = 0;
  }
#endif

  netif->state = (void*) 666;         /* !! */
  netif_set_link_up (netif);

  LWIP_TRACE (1, "eth_addr %02X:%02X:%02X:%02X:%02X:%02X\n",
             netif->hwaddr[0], netif->hwaddr[1], netif->hwaddr[2],
             netif->hwaddr[3], netif->hwaddr[4], netif->hwaddr[5]);

  LWIP_DEBUGF (PKT_IF_DEBUG | LWIP_DBG_TRACE, ("pkt_init\n"));

/*
 * From the C-preprocessor:
 * if ( ((1 | 0x40U) & 0x80U) &&
 *          ((1 | 0x40U) & lwip_debug_flags) &&
 *          (((1 | 0x40U) & 0x03) >= 0) ) {
 *             printf ("pkt_init\n");
 *             if ((1 | 0x40U) & 0x08U) { while(1); } } ;
 */

  return (ERR_OK);
}

/* Release the packet-driver.
 */
err_t
pkt_exit (struct netif *netif)
{
  if (PktGetStatistics(pktInfo.handle))
  {
    u32_t in_pkt   = pktStat.inPackets  - initial_stat.inPackets;
    u32_t out_pkt  = pktStat.outPackets - initial_stat.outPackets;
    u32_t in_err   = pktStat.inErrors   - initial_stat.inErrors;
    u32_t out_err  = pktStat.outErrors  - initial_stat.outErrors;
    u32_t in_lost  = pktStat.inLost     - initial_stat.inLost;

    printf ("pkt_stats: rx: %lu, tx: %lu, rx-err: %lu, tx-err: %lu, rx-lost: %lu, rx-dropped: %lu\n",
            in_pkt, out_pkt, in_err, out_err, in_lost, PktRxDropped());
  }
  else
    printf ("pkt_stats: err: %s\n", pktInfo.error);

  PktExitDriver();
  return (ERR_OK);
}

/*
 * Transmit a packet. The packet is contained in the pbuf that is passed to
 * the function. This pbuf might be chained.
 */
static err_t
low_level_output (struct netif *netif, struct pbuf *p)
{
  const struct eth_addr *dest;
  const struct eth_addr *src;
  const struct pbuf *q;

  u8_t   buffer [1514+10];
  u8_t  *ptr, *buf;
  u16_t  tot_len = p->tot_len;

#if defined(LWIP_DEBUG) && LWIP_NETIF_TX_SINGLE_PBUF
  LWIP_ASSERT ("p->next == NULL && p->len == p->tot_len", p->next == NULL && p->len == p->tot_len);
#endif

  /* initiate transfer */
  if (p->len == p->tot_len)
  {
    /* no pbuf chain, don't have to copy -> faster */
    buf = (u8_t*)p->payload;
  }
  else
  {
    /* pbuf chain, copy into contiguous buffer */
    if (p->tot_len >= sizeof(buffer))
    {
      LINK_STATS_INC (link.lenerr);
      LINK_STATS_INC (link.drop);
      snmp_inc_ifoutdiscards (netif);
      return (ERR_BUF);
    }
    buf = ptr = buffer;
    for (q = p; q; q = q->next)
    {
      /* Send the data from the pbuf to the interface, one pbuf at a
       * time. The size of the data in each pbuf is kept in 'q->len'.
       */
      LWIP_TRACE (4, "send ptr %p q->payload %p q->len %i q->next %p\n",
                  ptr, q->payload, (int)q->len, (void*)q->next);
      memcpy (ptr, q->payload, q->len);
      ptr += q->len;
    }
  }

  dest = (const struct eth_addr*) buf;
  src  = (const struct eth_addr*) dest + 1;

  LWIP_TRACE (4, "eth: src: %02X:%02X:%02X:%02X:%02X:%02X, dst: %02X:%02X:%02X:%02X:%02X:%02X\n",
             src->addr[0],  src->addr[1],  src->addr[2],
             src->addr[3],  src->addr[4],  src->addr[5],
             dest->addr[0], dest->addr[1], dest->addr[2],
             dest->addr[3], dest->addr[4], dest->addr[5]);

  /* signal that packet should be sent */
  if (!PktTransmit(buf, p->tot_len))
  {
    LINK_STATS_INC(link.memerr);
    LINK_STATS_INC(link.drop);
    snmp_inc_ifoutdiscards(netif);
    return ERR_BUF;
  }

  LINK_STATS_INC (link.xmit);
  snmp_add_ifoutoctets (netif, tot_len);

  if ((dest->addr[0] & 1) != 0)
  {
    /* broadcast or multicast packet*/
    MIB2_STATS_NETIF_INC (netif, ifoutnucastpkts);
  }
  else
  {
    /* unicast packet */
    MIB2_STATS_NETIF_INC (netif, ifoutucastpkts);
  }
  return (ERR_OK);
}


/*
 * Using NDIS3PKT or SwsVpkt causes all non-broadcast packets sent
 * to appear in receive buffer approx 100 msec later.
 */
static __inline int
is_looped (const struct netif *netif, const void *packet)
{
  const struct eth_hdr *eth = (const struct eth_hdr*) packet;
  const struct ip_hdr   *ip = (const struct ip_hdr*) eth + 1;
  const u32_t *src, *my_ip;

  if (memcmp(&eth->src, &netif->hwaddr, ETHARP_HWADDR_LEN))
     return (0);

  if (eth->type != 0x0800)    /* assume non-looped if not IPv4 */
     return (0);

  /* if eth->src matches, but IP-src is different it could be
   * Winsock sending. Thus it's not a looped packet.
   */
  src   = (const u32_t*) &ip->src;
  my_ip = (const u32_t*) &netif->ip_addr;

  return (*src == *my_ip);
}

/** low_level_input(): Allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 */
static
struct pbuf *low_level_input (struct netif *netif,
                              const char *packet, int packet_len,
                              err_t *err)
{
  const struct eth_addr *dest = (const struct eth_addr*) packet;
  const struct eth_addr *src = dest + 1;
  struct pbuf *p, *q;

  static const u8_t brdcast[]   = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
  static const u8_t ipv4mcast[] = { 0x01, 0x00, 0x5e };
  static const u8_t ipv6mcast[] = { 0x33, 0x33 };

  int from_us, to_us, is_brdcast;
  int is_unicast, is_ipv4_mcast, is_ipv6_mcast;

  LWIP_TRACE (2, "eth: src: %02X:%02X:%02X:%02X:%02X:%02X, dst: %02X:%02X:%02X:%02X:%02X:%02X\n",
             src->addr[0],  src->addr[1],  src->addr[2],
             src->addr[3],  src->addr[4],  src->addr[5],
             dest->addr[0], dest->addr[1], dest->addr[2],
             dest->addr[3], dest->addr[4], dest->addr[5]);

  to_us         = (memcmp(dest, &netif->hwaddr, ETHARP_HWADDR_LEN) == 0);
  from_us       = (memcmp(src, &netif->hwaddr, ETHARP_HWADDR_LEN) == 0);
  is_ipv4_mcast = (memcmp(dest, ipv4mcast, 3) == 0 && (dest->addr[3] & 0x80) == 0);
  is_ipv6_mcast = (memcmp(dest, ipv6mcast, 2) == 0);
  is_brdcast    = (memcmp(dest, brdcast, 6) == 0);
  is_unicast    = ((dest->addr[0] & 0x01) == 0);

  *err = ERR_OK;

  if (!to_us && !is_ipv4_mcast && !is_ipv6_mcast && !is_brdcast)
     goto drop_it;

  /*
   * Hack section: if transmitting to this lwIP program (running in a DOS-box under Windows)
   * from a Windows program, obviously the 'from_us' will be true (since changing our MAC-
   * address will surely fail). But the packet is not really from us. LwIP should never
   * send anything causing 'from_us' and 'to_us' be both be true unless it's a looped
   * packet. Hence if 'from_us && !is_looped()', let it in.
   */
#if 0
  if (!(from_us && !is_looped(netif,packet)))
    goto drop_it;
#endif

  /* We allocate a pbuf chain of pbufs from the pool. */
  p = pbuf_alloc (PBUF_RAW, (u16_t)packet_len, PBUF_POOL);
  if (p)
  {
    /* We iterate over the pbuf chain until we have read the entire
     * packet into the pbuf.
     */
    int start = 0;

    for (q = p; q; q = q->next)
    {
      u16_t copy_len = q->len;

      /* Read enough bytes to fill this pbuf in the chain. The
       * available data in the pbuf is given by 'q->len'.
       */
      LWIP_TRACE (4, "recv start %i packet_len %i q->payload %p "
                     "q->len %i q->next %p\n",
                 start, packet_len, q->payload, (int)q->len, (void*)q->next);
      memcpy (q->payload, packet + start, copy_len);
      start      += copy_len;
      packet_len -= copy_len;
      if (packet_len <= 0)
         break;
    }
    LINK_STATS_INC (link.recv);
    MIB2_STATS_NETIF_ADD(netif, ifinoctets, p->tot_len);
    if (is_unicast)
         MIB2_STATS_NETIF_INC (netif, ifinucastpkts);
    else MIB2_STATS_NETIF_INC (netif, ifinnucastpkts);
  }
  else
  {
    /* drop packet(); */
    LINK_STATS_INC (link.memerr);
    LINK_STATS_INC (link.drop);
    *err = -1;
  }
  return (p);

drop_it:

  /* don't update counters here! */
  LWIP_TRACE (2, "dropping pkt; from_us %d, to_us %d, is_ipv4_mcast %d, is_ipv6_mcast %d, is_brdcast %d\n",
              from_us, to_us, is_ipv4_mcast, is_ipv6_mcast, is_brdcast);
  *err = -1;
  return (NULL);
}

/* Poll the network-interface and process all input.
 */
err_t
pkt_poll (struct netif *netif)
{
  struct pbuf *p;
  char   buf [1514];
  int    len = PktReceive (buf, sizeof(buf));
  err_t  err = ERR_OK;

  if (len <= 0)
     return (err);

  LWIP_TRACE (4, "len: %d\n", len);

  /* move received packet into a new pbuf */
  p = low_level_input (netif, buf, len, &err);
  if (!p)
     return (err);

  /* pass all packets to ethernet_input, which decides what packets it supports */
  err = (*netif->input) (p, netif);
  if (err != ERR_OK)
  {
    LWIP_TRACE (1, "IP input error\n");
    pbuf_free (p);
  }
  return (err);
}

