#ifndef LWIP_PKTDRVR_H
#define LWIP_PKTDRVR_H

#ifndef LWIP_ARCH_CC_H
#error "Include <lwip.h> before this file."
#endif

#include "netif/etharp.h"

typedef enum PKT_CLASS {       /* Packet-driver classes */
        PD_ETHER      = 1,
        PD_PRONET10   = 2,
        PD_IEEE8025   = 3,
        PD_OMNINET    = 4,
        PD_APPLETALK  = 5,
        PD_SLIP       = 6,
        PD_STARTLAN   = 7,
        PD_ARCNET     = 8,
        PD_AX25       = 9,
        PD_KISS       = 10,
        PD_IEEE8023_2 = 11,
        PD_FDDI8022   = 12,
        PD_X25        = 13,
        PD_LANstar    = 14,
        PD_PPP        = 18
      } PKT_CLASS;

typedef enum PKT_RX_MODE {  /* Packet-driver receive modes    */
        PDRX_OFF    = 1,    /* turn off receiver              */
        PDRX_DIRECT,        /* receive only to this interface */
        PDRX_BROADCAST,     /* DIRECT + broadcast packets     */
        PDRX_MULTICAST1,    /* BROADCAST + limited multicast  */
        PDRX_MULTICAST2,    /* BROADCAST + all multicast      */
        PDRX_ALL_PACKETS,   /* receive all packets on network */
      } PKT_RX_MODE;

#include <arch/bpstruct.h>

typedef struct PKT_INFO {
        u8_t  class;        /* = 1 for DEC/Intel/Xerox Ethernet  */
        u8_t  number;       /* = 0 for single LAN adapter        */
        u16_t type;         /* = 13 for 3C523                    */
        u8_t  funcs;        /* Basic/Extended/HiPerf functions   */
        u16_t intr;         /* user interrupt vector number      */
        u16_t handle;       /* Handle associated with session    */
        u8_t  name [15];    /* Name of adapter interface, ie. "3C523" */
        int   quiet;        /* (don't) print errors to stdout    */
        const char *error;  /* address of error string           */
        u8_t  majVer;       /* Major driver implementation ver.  */
        u8_t  minVer;       /* Minor driver implementation ver.  */
        u8_t  dummyLen;     /* length of following data          */
        u16_t MAClength;    /* HiPerformance data, N/A           */
        u16_t MTU;          /* HiPerformance data, N/A           */
        u16_t multicast;    /* HiPerformance data, N/A           */
        u16_t rcvrBuffers;  /* valid for                         */
        u16_t UMTbufs;      /*   High Performance drivers only   */
        u16_t postEOIintr;  /*                  Usage ??         */
      } PKT_INFO;

#define PKT_PARAM_SIZE  14    /* members majVer - postEOIintr */

typedef struct PKT_STAT {
        u32_t inPackets;          /* # of packets received    */
        u32_t outPackets;         /* # of packets transmitted */
        u32_t inBytes;            /* # of bytes received      */
        u32_t outBytes;           /* # of bytes transmitted   */
        u32_t inErrors;           /* # of reception errors    */
        u32_t outErrors;          /* # of transmission errors */
        u32_t inLost;             /* # of packets lost on RX  */
      } PKT_STAT;

#include <arch/epstruct.h>

#define PKT_ERR_BUF_SIZE 256

/*
 * Packet-driver instance data. All information for one handle is kept in
 * this structure. The caller of below function must allocate and clear
 * it before calling *any* of the below Pkt*() functions.
 */
struct pkt_info2 {
       struct PKT_INFO    info;
       struct PKT_STAT    initial_stat;
       struct PKT_STAT    last_stat;
       enum   PKT_RX_MODE receive_mode;
       int                debug;
       u8_t               my_address [ETHARP_HWADDR_LEN];
       u8_t               broadcast  [ETHARP_HWADDR_LEN];
       char               last_error [PKT_ERR_BUF_SIZE];

#if defined(__DJGPP__)
       _go32_dpmi_seginfo rm_mem;
       u32_t              realBase;

#elif defined(__WATCOMC__)
       u16_t               rm_base_seg, rm_base_sel;
       u32_t               realBase;
#endif

       struct {                /* internal statistics: */
         u32_t  too_small;     /* size < 60 */
         u32_t  too_large;     /* size > 1514 */
         u32_t  bad_sync;      /* count_1 != count_2 */
         u32_t  wrong_handle;  /* upcall to wrong handle */
       } int_stat;

     } PKT_INFO2;

extern PKT_STAT    pktStat;     /* statistics for packets */
extern PKT_INFO    pktInfo;     /* packet-driver information */

extern PKT_RX_MODE     receiveMode;
extern struct eth_addr myAddress;

extern int  PktInitDriver (PKT_RX_MODE mode);
extern int  PktExitDriver (void);

extern const char *PktGetErrorStr    (int errNum);
extern const char *PktGetClassName   (u16_t class);
extern const char *PktRXmodeStr      (PKT_RX_MODE mode);
extern int         PktSearchDriver   (void);
extern int         PktReceive        (char *buf, int max);
extern int         PktTransmit       (const void *eth, int len);
extern int         PktReleaseHandle  (u16_t handle);
extern int         PktTerminHandle   (u16_t handle);
extern int         PktResetInterface (u16_t handle);
extern int         PktSetReceiverMode(PKT_RX_MODE  mode);
extern int         PktGetReceiverMode(PKT_RX_MODE *mode);
extern int         PktGetStatistics  (u16_t handle);
extern int         PktGetAddress     (struct eth_addr *addr);
extern int         PktSetAddress     (const struct eth_addr *addr);
extern int         PktGetDriverInfo  (void);
extern int         PktGetDriverParam (void);
extern void        PktQueueBusy      (int busy);
extern u16_t       PktBuffersUsed    (void);
extern u32_t       PktRxDropped      (void);

#endif /* LWIP_PKTDRVR_H */

